from typing import Any

from rebyu.util.dependency import nltk_dependency_mgt

from nltk.sentiment.vader import SentimentIntensityAnalyzer
from textblob import TextBlob


def textblob_polarity(series: Any):
    """ Predict the polarity of a given text data using TextBlob

    :param series: Any series of data
    :return: List of Dict (TextBlob Polarity)
    """
    nltk_dependency_mgt(required=['punkt'])

    polarities = []
    for text in series:
        if isinstance(text, TextBlob):
            polarities.append(text.polarity)
        else:
            polarities.append(TextBlob(text).polarity)
    return polarities


def vader_polarity(series: Any):
    """ Predict the polarity of a given text data using VADER

    :param series: Any series of data
    :return: List of Dict (Vader Polarity)
    """
    polarities = []
    analyzer = SentimentIntensityAnalyzer()
    for text in series:
        if isinstance(text, TextBlob):
            polarities.append(analyzer.polarity_scores(text.raw))
        else:
            polarities.append(analyzer.polarity_scores(text))
    return polarities




import sys

from typing import AnyStr

from loguru import logger


class Logger(object):

    def __init__(self, class_name):
        self.class_name = class_name
        self.ctx_log = logger.bind(name=class_name)

    def info(self, message: AnyStr):
        self.ctx_log.info(message)

    def warning(self, message: AnyStr):
        self.ctx_log.warning(message)

    def error(self, message: AnyStr):
        self.ctx_log.error(message)



import pathlib
from typing import Any, Dict, AnyStr

import pandas as pd

from rebyu.pipeline.pipeline import BasePipeline


class BaseRebyu(object):
    """
    Base Class for Rebyu

    The BaseRebyu class is the foundation of Rebyu, an automatic review analysis pipeline and toolkit.
    It processes various types of data, primarily text data, using built-in pipelines designed for
    review analysis. Rebyu handles preprocessing, composition extraction, and analysis using its own
    tools and external toolsets.

    Attributes:
        data (pd.DataFrame): The input dataset, converted to a DataFrame.
        pipeline (BasePipeline): A customizable pipeline for data preprocessing, composition,
            and analysis.
        verbose (bool): Determines whether verbose output is enabled.
    """

    def __init__(self,
                 data: Any,
                 pipeline: BasePipeline,
                 verbose: bool = True):
        self.data = data
        self.pipeline = None
        self._composition = {}
        self._analysis = {}

        self.verbose = verbose

        if isinstance(data, pd.DataFrame):
            self.data = data.copy()

        if isinstance(data, pd.Series):
            self.data = self.data.to_frame().copy()

        if isinstance(data, str):
            _suffix = pathlib.Path(data).suffix

            if _suffix == '.csv':
                self.data = pd.read_csv(data)
            if _suffix == '.json':
                self.data = pd.read_json(data)

        if isinstance(pipeline, BasePipeline):
            self.pipeline = pipeline
            self.pipeline.reset()

    def step(self, verbose: bool = False):
        """ Take a step of a series of process within the pipeline.

        :param verbose: Determines whether verbose output is enabled.
        :return:
        """
        if self.pipeline is None:
            return

        self.pipeline.step(
            data=self.data,
            composition=self._composition,
            analysis=self._analysis,
            verbose=verbose
        )

    def run(self, verbose: bool = False):
        """ Run through all the steps within the pipeline.

        :param verbose: Determines whether verbose output is enabled.
        :return:
        """
        if self.pipeline is None:
            return

        self.pipeline.run(
            data=self.data,
            composition=self._composition,
            analysis=self._analysis,
            verbose=verbose
        )

    def composition(self, key: AnyStr, default: Any = None) -> Any:
        """ Get values from 'composition' step results

        :param key: Target Key of Composition
        :param default: (Optional) Default value if not found
        :return: Any
        """
        return self._composition.get(key, default)

    def analysis(self, key: AnyStr, default: Any = None) -> Any:
        """ Get values from 'analysis' step results

        :param key: Target Key of Analysis
        :param default: (Optional) Default value if not found
        :return: Any
        """
        return self._analysis.get(key, default)

    def reset(self):
        """ Reset the results of Rebyu object. (Composition Result, Analysis Result, and Pipeline State)

        :return:
        """
        self._composition.clear()
        self._analysis.clear()
        self.pipeline.reset()

    def info(self):
        """ Get the information of the class.

        :return:
        """
        out_text = 'DATA:\n'
        for col in self.data.columns:
            out_text += f' -{col}: {self.data[col].dtype}\n'

        out_text += '\nCOMPOSITION:\n'
        if len(self._composition):
            for kc, vc in self._composition.items():
                out_text += f' -{kc}: {type(vc)}\n'
        else:
            out_text += 'Empty\n'

        out_text += '\nANALYSIS:\n'
        if len(self._analysis):
            for kc, vc in self._analysis.items():
                out_text += f' -{kc}: {type(vc)}\n'
        else:
            out_text += 'Empty\n'

        out_text += '\nPIPELINE:\n'
        out_text += f'pid: {self.pipeline.pid}\n'
        out_text += f'steps: {len(self.pipeline)}\n'
        out_text += f'state: {self.pipeline.state()}\n'

        out_text += '\nSTEPS:\n'
        steps_info = self.pipeline.steps_info()
        for step in steps_info:
            out_text += f' -{step}\n'

        return out_text



import time

from typing import Any, Dict, List

from rebyu.base import BaseRebyu
from rebyu.pipeline.base import BasePipeline
from rebyu.pipeline import BLANK_PIPELINE


class Rebyu(BaseRebyu):
    """
    Rebyu

    The main class to access the powerful and customizable toolkit for Review Analysis.
    Rebyu takes the approach of a continuous pipeline operating under 3 categories (preprocess,
    composition, and analysis).

    Read More https://github.com/abhishtagatya/rebyu
    """

    def __init__(self,
                 data: Any,
                 pipeline: BasePipeline = BLANK_PIPELINE):
        super(Rebyu, self).__init__(data, pipeline)



import re
import string
from typing import List, Any, AnyStr

import nltk
from nltk.corpus import stopwords

from rebyu.util.dependency import nltk_dependency_mgt

NUMBERS_REGEX = r'\d+'
URLS_REGEX = r'(https:\/\/www\.|http:\/\/www\.|https:\/\/|http:\/\/)?[a-zA-Z]{2,}(\.[a-zA-Z]{2,})(\.[a-zA-Z]{2,' \
             r'})?\/[a-zA-Z0-9]{2,}|((https:\/\/www\.|http:\/\/www\.|https:\/\/|http:\/\/)?[a-zA-Z]{2,}(\.[a-zA-Z]{2,' \
             r'})(\.[a-zA-Z]{2,})?)|(https:\/\/www\.|http:\/\/www\.|https:\/\/|http:\/\/)?[a-zA-Z0-9]{2,' \
             r'}\.[a-zA-Z0-9]{2,}\.[a-zA-Z0-9]{2,}(\.[a-zA-Z0-9]{2,})? '

EMOJI_REGEX = re.compile("["
                           u"\U0001F600-\U0001F64F"  # emoticons
                           u"\U0001F300-\U0001F5FF"  # symbols & pictographs
                           u"\U0001F680-\U0001F6FF"  # transport & map symbols
                           u"\U0001F700-\U0001F77F"  # alchemical symbols
                           u"\U0001F780-\U0001F7FF"  # Geometric Shapes Extended
                           u"\U0001F800-\U0001F8FF"  # Supplemental Arrows-C
                           u"\U0001F900-\U0001F9FF"  # Supplemental Symbols and Pictographs
                           u"\U0001FA00-\U0001FA6F"  # Chess Symbols
                           u"\U0001FA70-\U0001FAFF"  # Symbols and Pictographs Extended-A
                           u"\U0000FE00-\U0000FE0F"  # Variation Selectors
                           u"\U00002702-\U000027B0"  # Dingbats
                           "]+", flags=re.UNICODE)


def remove_patterns(text: Any, pattern: AnyStr):
    """ Remove patterns (regex) from a text

    :param text: Any string object
    :param pattern: Any string object
    :return: text
    """
    if type(text) is str:
        return re.sub(pattern, '', text)
    return text


def remove_numbers(text: Any, pattern: AnyStr = NUMBERS_REGEX):
    """ Remove numerical data from text (using a regex pattern)

    :param text: Any string object
    :param pattern: Any string object
    :return: text
    """
    return remove_patterns(text, pattern)


def remove_urls(text: Any, pattern: AnyStr = URLS_REGEX):
    """ Remove urls from text (using a regex pattern)

    :param text: Any string object
    :param pattern: Any string object
    :return: text
    """
    return remove_patterns(text, pattern)


def remove_emojis(text: Any, pattern: AnyStr = EMOJI_REGEX):
    """ Remove emojis from text
    (Emoticons, Symbols & Pictograms, Transport & Map Symbols, Alchemical Symbols,
    Geometric Shape Extended, Supplemental Arrows-C, Supplemental Symbols & Pictographs,
    Chess Symbols, Symbols and Pictographs Extended-A, Dingbats, and Enclosed Alphanumerics & Ideographic)

    :param text: Any string object
    :param pattern: Any string object
    :return: text
    """
    return remove_patterns(text, pattern)


def remove_punctuations(text: Any):
    """ Remove punctuations from text (from string.punctuation)

    :param text: Any string object
    :return: text
    """
    if type(text) is str:
        return text.translate(str.maketrans('', '', string.punctuation))
    return text


def remove_whitespaces(text: Any):
    """ Strip any whitespaces from text

    :param text: Any string object
    :return: text
    """
    if type(text) is str:
        return text.strip()
    return text


def remove_specifics(text: Any, sub: Any):
    """ Remove specific substring from text

    :param text: Any string object
    :param sub: Any string object
    :return: text
    """
    if type(text) is str and type(sub) is str:
        return text.replace(sub, '')
    return text


def trim_text(text: Any, length: int = 500, pos: AnyStr = 'LEFT'):
    """ Trim the Text by Length

    :param text: Any string object
    :param length: Max. String length
    :param pos: Trim Position (LEFT or RIGHT)
    :return: text
    """
    if pos == 'LEFT':
        return text[:length]
    return text[length:]


# -- NLTK -- #

def remove_stopwords(text: Any, extra: List[Any] = None, language: str = 'english'):
    """ Remove stopwords from text (using NLTK Stopwords)

    :param text: Any string object
    :param extra: List of Any string objects
    :param language: The language of the text
    :return: text
    """

    nltk_dependency_mgt(required=['punkt', 'stopwords'])

    stop_words = set(stopwords.words(language) + (extra or []))
    tokens = nltk.word_tokenize(text)
    filtered_tokens = filter(lambda x: x.lower() not in stop_words, tokens)

    if type(text) is str:
        return ' '.join(filtered_tokens)
    return filtered_tokens



from typing import Any, AnyStr, Dict

from rebyu.pipeline.base import BaseStep
from rebyu.preprocess.remove import (
    remove_patterns,
    remove_numbers,
    remove_punctuations,
    remove_whitespaces,
    remove_specifics,
    remove_urls,
    remove_emojis,
    remove_stopwords,
    trim_text
)
from rebyu.preprocess.transform import (
    cast_nan_str,
    cast_case,
    sub_replace,
    expand_contractions,
    censor_username,
    censor_urls,

    # TextBlob
    to_textblob,
    textblob_tokenize,
    textblob_sentences,

    # NLTK
    nltk_tokenize,
    nltk_porter_stem,
    nltk_lancaster_stem,
    nltk_wordnet_lemma,

    # Transformers
    transformers_tokenize
)
from rebyu.compose.vocab import (
    counter_vocab,
    counter_character_vocab,
    set_character_vocab,

    # NLTK
    nltk_vocab
)
from rebyu.compose.ner import nltk_extract_ner
from rebyu.compose.pos import nltk_extract_pos_tags
from rebyu.analysis.sentiment import (
    # TextBlob
    textblob_polarity,

    # NLTK
    vader_polarity
)
from rebyu.analysis.misc import (
    transformers_model,
    transformers_pipeline
)


class RebyuStep(BaseStep):
    """
    RebyuStep

    The main class to access or initiate a Rebyu Step. A step is an encapsulated function to operate under a
    pipeline which Rebyu provides. The function takes a source of data (which would be a key from the Rebyu.data)
    DataFrame and a target to output its data. The behaviour of the operation may differ depending on the
    type of step it is. It can be PREPROCESS, COMPOSE, or ANALYZE.
    """

    def __init__(self,
                 sid: AnyStr,
                 stype: AnyStr,
                 source: Any,
                 target: Any,
                 func: Any,
                 func_args: Dict[AnyStr, Any] = None):
        super(RebyuStep, self).__init__(
            sid=sid,
            stype=stype,
            source=source,
            target=target,
            func=func,
            func_args=func_args
        )


PREP_CAST_NAN = RebyuStep(
    sid='rb-cast_nan',
    stype=RebyuStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=cast_nan_str
)
"""Convert non-string data in 'text' (Rebyu.data) to an empty string."""

PREP_CAST_CASE = RebyuStep(
    sid='rb-cast_case',
    stype=RebyuStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=cast_case
)
"""Switch case the data in 'text' (Rebyu.data) to lower or upper case."""

PREP_TRIM_TEXT = RebyuStep(
    sid='rb-trim_text',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=trim_text
)
"""Trim the text from 'text' to a given length"""

PREP_REMOVE_PATTERNS = RebyuStep(
    sid='rb-remove_patterns',
    stype=RebyuStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=remove_patterns
)
"""Remove numerical data in 'text' (Rebyu.data)"""

PREP_REMOVE_NUMBERS = RebyuStep(
    sid='rb-remove_numbers',
    stype=RebyuStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=remove_numbers
)
"""Remove numerical data in 'text' (Rebyu.data)"""

PREP_REMOVE_PUNCTUATIONS = RebyuStep(
    sid='rb-remove_punctuations',
    stype=RebyuStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=remove_punctuations
)
"""Remove punctuations in 'text' (Rebyu.data)"""

PREP_REMOVE_WHITESPACES = RebyuStep(
    sid='rb-remove_whitespaces',
    stype=RebyuStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=remove_whitespaces
)
"""Remove whitespaces in 'text' (Rebyu.data)"""

PREP_REMOVE_SPECIFICS = RebyuStep(
    sid='rb-remove_specifics',
    stype=RebyuStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=remove_specifics
)
"""Remove specific substring in 'text' (Rebyu.data)"""

PREP_REMOVE_URLS = RebyuStep(
    sid='rb-remove_urls',
    stype=RebyuStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=remove_urls
)
"""Remove urls in 'text' (Rebyu.data)"""

PREP_REMOVE_EMOJIS = RebyuStep(
    sid='rb-remove_emojis',
    stype=RebyuStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=remove_emojis
)
"""Remove emojis in 'text' (Rebyu.data)"""

PREP_REMOVE_STOPWORDS = RebyuStep(
    sid='rb-remove_stopwords',
    stype=RebyuStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=remove_stopwords
)
"""Remove stopwords in 'text' (Rebyu.data)"""

PREP_REPLACE_WORD = RebyuStep(
    sid='rb-sub_replace',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=sub_replace
)
"""Replace substring in 'text' (Rebyu.data)"""

PREP_EXPAND_CONTRACTIONS = RebyuStep(
    sid='rb-expand_contractions',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=expand_contractions
)
"""Replace contractions in 'text' (Rebyu.data) to full words"""

PREP_CENSOR_USERNAME = RebyuStep(
    sid='rb-censor_username',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=censor_username
)
"""Censor any username in 'text' (Rebyu.data) to a placeholder"""

PREP_CENSOR_URLS = RebyuStep(
    sid='rb-censor_urls',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=censor_urls
)
"""Censor any urls in 'text' (Rebyu.data) to a placeholder"""

PREP_SENTENCE_LENGTH = RebyuStep(
    sid='rb-sentence_length',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='length',
    func=lambda x: len(x) if type(x) is str else ''
)
"""Return the sentence length of 'text' (Rebyu.data) to 'length'"""

PREP_WORD_COUNT = RebyuStep(
    sid='rb-word_count',
    stype=BaseStep.STEP_PREPROCESS,
    source='tokens',
    target='word_counts',
    func=lambda x: len(x)
)
"""Return the word count of 'text' (Rebyu.data) to 'word_counts'"""

COMPOSE_COUNTER_VOCAB = RebyuStep(
    sid='rb-counter_vocab',
    stype=BaseStep.STEP_COMPOSE,
    source='tokens',
    target='vocab',
    func=counter_vocab
)
"""Compose a vocabulary from 'tokens' (Rebyu.data) to 'vocab' using Counter from collections"""

COMPOSE_COUNTER_CHARVOCAB = RebyuStep(
    sid='rb-counter_charvocab',
    stype=BaseStep.STEP_COMPOSE,
    source='tokens',
    target='char_vocab',
    func=counter_character_vocab
)
"""Compose a character set from 'tokens' (Rebyu.data) to 'char_vocab' using Counter from collections"""

COMPOSE_SET_CHARVOCAB = RebyuStep(
    sid='rb-set_charvocab',
    stype=BaseStep.STEP_COMPOSE,
    source='tokens',
    target='char_vocab',
    func=set_character_vocab
)
"""Compose a character set from 'tokens' (Rebyu.data) to 'char_vocab' using a built-in set"""


# -- TextBlob -- #

PREP_TEXTBLOB = RebyuStep(
    sid='rb-to_textblob',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=to_textblob
)
"""Converts data into a TextBlob object from 'text' (Rebyu.data)"""

PREP_TEXTBLOB_TOKENIZE = RebyuStep(
    sid='rb-textblob_tokenize',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='tokens',
    func=textblob_tokenize
)
"""Tokenize 'text' (Rebyu.data) into 'tokens' using `TextBlob.tokens`"""

PREP_TEXTBLOB_SENTENCES = RebyuStep(
    sid='rb-textblob_sentences',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='sentences',
    func=textblob_sentences
)
"""Extract sentences from 'text' (Rebyu.data) into 'sentences' using `TextBlob.sentences`"""

ANALYZE_TEXTBLOB_POLARITY = RebyuStep(
    sid='rb-textblob_polarity',
    stype=BaseStep.STEP_ANALYZE,
    source='text',
    target='textblob_polarity',
    func=textblob_polarity
)
"""Predict the polarity of 'text' (Rebyu.data) into 'textblob_polarity' using `TextBlob.polarity`"""

# -- NLTK -- #

PREP_NLTK_TOKENIZE = RebyuStep(
    sid='rb-tokenize',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='tokens',
    func=nltk_tokenize
)
"""Tokenize 'text' (Rebyu.data) into 'tokens' using `nltk.word_tokenize`"""

PREP_NLTK_PORTER_STEM = RebyuStep(
    sid='rb-porter_stem',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=nltk_porter_stem
)
"""Stem each word from 'text' (Rebyu.data) using the Porter method from NLTK"""

PREP_NLTK_LANCASTER_STEM = RebyuStep(
    sid='rb-porter_stem',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=nltk_lancaster_stem
)
"""Stem each word from 'text' (Rebyu.data) using the Lancaster method from NLTK"""

PREP_NLTK_WORDNET_LEMMA = RebyuStep(
    sid='rb-wordnet_lemma',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='text',
    func=nltk_wordnet_lemma
)
"""Lemmatize each word from 'text' (Rebyu.data) using WordNET from NLTK"""

COMPOSE_NLTK_VOCAB = RebyuStep(
    sid='rb-nltk_vocab',
    stype=BaseStep.STEP_COMPOSE,
    source='tokens',
    target='vocab',
    func=nltk_vocab
)
"""Compose a vocabulary 'tokens' (Rebyu.data) using NLTK's Vocabulary class"""

COMPOSE_NLTK_POS_TAG = RebyuStep(
    sid='rb-nltk_pos_tag',
    stype=BaseStep.STEP_COMPOSE,
    source='tokens',
    target='pos_tags',
    func=nltk_extract_pos_tags
)
"""Extract Part-of-Speech Tags from 'tokens' (Rebyu.data) using NLTK's pos_tag"""

COMPOSE_NLTK_NER = RebyuStep(
    sid='rb-nltk_ner',
    stype=BaseStep.STEP_COMPOSE,
    source='tokens',
    target='ner',
    func=nltk_extract_ner
)
"""Extract Named Entities from 'tokens' (Rebyu.data) using NLTK's ne_chunk"""

ANALYZE_VADER_POLARITY = RebyuStep(
    sid='rb-vader_polarity',
    stype=BaseStep.STEP_ANALYZE,
    source='text',
    target='vader_polarity',
    func=vader_polarity
)
"""Predict the polarity of 'text' (Rebyu.data) into 'vader_polarity' using the VADER algorithm"""

# -- Transformers -- #

PREP_TRANSFORMERS_TOKENIZE = RebyuStep(
    sid='rb-transformers_tokenize',
    stype=BaseStep.STEP_PREPROCESS,
    source='text',
    target='tokens',
    func=transformers_tokenize
)
"""Tokenize 'text' (Rebyu.data) into 'tokens' using transformers module tokenizers

See the `transformers_tokenize` documentation."""

ANALYZE_TRANSFORMERS_MODEL = RebyuStep(
    sid='rb-transformers_model',
    stype=BaseStep.STEP_ANALYZE,
    source='tokens',
    target='transformers_analysis',
    func=transformers_model
)
"""Predict values from 'tokens' (Rebyu.data) into an analysis in 'transformers_analysis'. It uses models from the 
transformers module. 

See the `transformers_model` documentation."""

ANALYZE_TRANSFORMERS_PIPELINE = RebyuStep(
    sid='rb-transformers_pipeline',
    stype=BaseStep.STEP_ANALYZE,
    source='text',
    target='transformers_analysis',
    func=transformers_pipeline
)
"""Construct a pipeline from the transformers module. Taking 'text' (Rebyu.data) into an analysis in 
'transformers_analysis' 

See the `transformers_pipeline` documentation."""

ANALYZE_CARDIFFNLP_SENTIMENT = RebyuStep(
    sid='cardiffnlp_sentiment',
    stype=BaseStep.STEP_ANALYZE,
    source='text',
    target='sentiment',
    func=transformers_pipeline,
    func_args={
        'task': 'text-classification',
        'model': 'cardiffnlp/twitter-roberta-base-sentiment-latest'
    }
)
"""Construct a pipeline from the transformers module specific for Cardiff NLP Sentiment Analysis Tasks.
Taking 'text' (Rebyu.data) into an analysis in 'sentiment' 

See the `transformers_pipeline` documentation."""

ANALYZE_CARDIFFNLP_EMOTION = RebyuStep(
    sid='cardiffnlp_emotion',
    stype=BaseStep.STEP_ANALYZE,
    source='text',
    target='emotion',
    func=transformers_pipeline,
    func_args={
        'task': 'text-classification',
        'model': 'cardiffnlp/twitter-roberta-base-emotion-multilabel-latest',
        'top_k': 10
    }
)
"""Construct a pipeline from the transformers module specific for Cardiff NLP Emotion Analysis Tasks.
Taking 'text' (Rebyu.data) into an analysis in 'emotion' 

See the `transformers_pipeline` documentation."""

ANALYZE_FINITEAUTOMATA_SENTIMENT = RebyuStep(
    sid='finiteautomata_sentiment',
    stype=BaseStep.STEP_ANALYZE,
    source='text',
    target='sentiment',
    func=transformers_pipeline,
    func_args={
        'task': 'text-classification',
        'model': 'finiteautomata/bertweet-base-sentiment-analysis'
    }
)
"""Construct a pipeline from the transformers module specific for @finiteautomata Sentiment Analysis Tasks.
Taking 'text' (Rebyu.data) into an analysis in 'sentiment' 

See the `transformers_pipeline` documentation."""

ANALYZE_FINITEAUTOMATA_EMOTION = RebyuStep(
    sid='finiteautomata_emotion',
    stype=BaseStep.STEP_ANALYZE,
    source='text',
    target='emotion',
    func=transformers_pipeline,
    func_args={
        'task': 'text-classification',
        'model': 'finiteautomata/bertweet-base-emotion-analysis',
        'top_k': 5
    }
)
"""Construct a pipeline from the transformers module specific for @finiteautomata Emotion Analysis Tasks.
Taking 'text' (Rebyu.data) into an analysis in 'emotion' 

See the `transformers_pipeline` documentation."""


from typing import List, Any, AnyStr
from rebyu.util.dependency import nltk_dependency_mgt

import nltk
from nltk.stem import PorterStemmer
from nltk.stem import LancasterStemmer
from nltk.stem import WordNetLemmatizer

from textblob import TextBlob
from transformers import AutoTokenizer

import contractions


def cast_nan_str(text: Any):
    """ Cast non-string objects into empty strings

    :param text: Any string object
    :return: text
    """
    return text if type(text) is str else ''


def cast_case(text: Any, case: AnyStr = 'LOWER'):
    """ Cast the case into a string object (upper or lowercase)

    :param text: Any string object
    :param case: LOWER or UPPER
    :return: text
    """

    if type(text) is str:
        if case.upper() == 'LOWER':
            return text.lower()
        if case.upper() == 'UPPER':
            return text.upper()
        return text

    new_token = []
    for token in text:
        if case.upper() == 'LOWER':
            new_token.append(token.lower())
        elif case.upper() == 'UPPER':
            new_token.append(token.upper())
        else:
            new_token.append(token)
    return new_token


def sub_replace(text: Any, sub: Any = None, rep: Any = None):
    """ Replace a substring with another string (sub -> rep)

    :param text: Any string object
    :param sub: Any string object
    :param rep: Any string object
    :return: text
    """
    if sub is rep:
        return text

    if type(text) is str:
        return text.replace(sub, rep)

    return [x if x != sub else rep for x in text]


def expand_contractions(text: Any):
    """ Expand the contractions within the text (you're -> you are)

    :param text: Any string object
    :return: text
    """
    nltk_dependency_mgt(required=['punkt'])

    if type(text) is str:
        return contractions.fix(text)

    expanded = []
    for token in text:
        for word in contractions.fix(token).split():
            expanded.append(word)
    return expanded


def censor_username(text: Any, censor: Any = '@user'):
    """ Censor any username within the text with a placeholder

    :param text: Any string object
    :param censor: Any string object (placeholder)
    :return: text
    """
    nltk_dependency_mgt(required=['punkt'])

    new_text = []
    if type(text) is str:
        for token in text.split():
            token = censor if token.startswith('@') and len(token) > 1 else token
            new_text.append(token)
        return ' '.join(new_text)

    for token in text:
        token = censor if token.startswith('@') and len(token) > 1 else token
        new_text.append(token)
    return new_text


def censor_urls(text: Any, censor: Any = 'http'):
    """ Censor any urls within the text with a placeholder

    :param text: Any string object
    :param censor: Any string object (placeholder)
    :return: text
    """
    nltk_dependency_mgt(required=['punkt'])

    new_text = []
    if type(text) is str:
        for token in text.split():
            token = censor if token.startswith('http') else token
            new_text.append(token)
        return ' '.join(new_text)

    for token in text:
        token = censor if token.startswith('http') else token
        new_text.append(token)
    return new_text


# -- TextBlob -- #

def to_textblob(text: Any):
    """ Transform text into a TextBlob Object

    :param text: Any string object
    :return: TextBlob
    """
    if isinstance(text, TextBlob):
        return text

    return TextBlob(text)


def textblob_tokenize(text: Any):
    """ Tokenize text using TextBlob tokens

    :param text: Any string object
    :return: List of Strings
    """
    if isinstance(text, TextBlob):
        return text.tokens

    return TextBlob(text).tokens


def textblob_sentences(text: Any):
    """ Extract the sentences using TextBlob sentences

    :param text: Any string object
    :return: List of Strings
    """
    if isinstance(text, TextBlob):
        return text.sentences

    return TextBlob(text).sentences


# -- NLTK -- #

def nltk_tokenize(text: Any, language: AnyStr = 'english'):
    """ Tokenize text using NLTK word_tokenize

    :param text: Any string object
    :param language: The language of the text
    :return: List of String
    """
    nltk_dependency_mgt(required=['punkt'])

    if type(text) is str:
        return nltk.word_tokenize(
            text, language=language
        )
    return text


def nltk_porter_stem(text: Any):
    """ Stem the words from text using Porter method

    :param text: Any string object
    :return: text
    """
    nltk_dependency_mgt(required=['punkt'])

    stemmer = PorterStemmer()
    if type(text) is str:
        return ' '.join([stemmer.stem(x) for x in nltk.word_tokenize(text)])
    return [stemmer.stem(x) for x in text]


def nltk_lancaster_stem(text: Any):
    """ Stem the words from text using Lancaster method

    :param text: Any string object
    :return: text
    """
    nltk_dependency_mgt(required=['punkt'])

    stemmer = LancasterStemmer()
    if type(text) is str:
        return ' '.join([stemmer.stem(x) for x in nltk.word_tokenize(text)])
    return [stemmer.stem(x) for x in text]


def nltk_wordnet_lemma(text: Any):
    """ Lemmatize the words from text using WordNet

    :param text: Any string object
    :return: text
    """
    nltk_dependency_mgt(required=['punkt', 'wordnet', 'averaged_perceptron_tagger'])

    lemma = WordNetLemmatizer()
    if type(text) is str:
        return ' '.join([
            lemma.lemmatize(x, pos=_lemma_penn2morphy(tag))
            for x, tag in nltk.pos_tag(nltk.word_tokenize(text))
        ])
    return [
        lemma.lemmatize(x, pos=_lemma_penn2morphy(tag))
        for x, tag in nltk.pos_tag(text)
    ]


def _lemma_penn2morphy(penn_tag: AnyStr):
    try:
        morphy_tag = {'NN': 'n', 'JJ': 'a', 'VB': 'v', 'RB': 'r'}
        return morphy_tag[penn_tag[:2]]
    except KeyError:
        return 'v'
    except IndexError:
        return 'v'


# -- Transformers -- #

def transformers_tokenize(
        text: Any,
        tokenizer: Any = AutoTokenizer,
        model_name: AnyStr = 'bert-base-uncased',
        returns: AnyStr = 'pt'):
    """ Tokenize text using transformers module

    :param text: Any string object
    :param tokenizer: The tokenizer model (from transformers)
    :param model_name: The pretrained model name
    :param returns: The return value (pt: PyTorch, tf: Tensorflow)
    :return: Tokenized Object
    """
    tokenizer = tokenizer.from_pretrained(model_name)
    return tokenizer(text, return_tensors=returns)



from rebyu.pipeline.base import BaseStep, BasePipeline
from rebyu.pipeline.step import (
    RebyuStep,
    PREP_CAST_NAN,
    PREP_CAST_CASE,
    PREP_TRIM_TEXT,
    PREP_REMOVE_PATTERNS,
    PREP_REMOVE_NUMBERS,
    PREP_REMOVE_PUNCTUATIONS,
    PREP_REMOVE_WHITESPACES,
    PREP_REMOVE_SPECIFICS,
    PREP_REMOVE_EMOJIS,
    PREP_REMOVE_STOPWORDS,
    PREP_REPLACE_WORD,
    PREP_CENSOR_USERNAME,
    PREP_CENSOR_URLS,
    PREP_EXPAND_CONTRACTIONS,
    PREP_SENTENCE_LENGTH,
    PREP_WORD_COUNT,
    COMPOSE_COUNTER_VOCAB,
    COMPOSE_COUNTER_CHARVOCAB,
    COMPOSE_SET_CHARVOCAB,

    # TextBlob
    PREP_TEXTBLOB,
    PREP_TEXTBLOB_TOKENIZE,
    PREP_TEXTBLOB_SENTENCES,
    ANALYZE_TEXTBLOB_POLARITY,

    # NLTK
    PREP_NLTK_TOKENIZE,
    COMPOSE_NLTK_VOCAB,
    COMPOSE_NLTK_POS_TAG,
    COMPOSE_NLTK_NER,
    ANALYZE_VADER_POLARITY,

    # Transformers
    PREP_TRANSFORMERS_TOKENIZE,
    ANALYZE_TRANSFORMERS_MODEL,
    ANALYZE_TRANSFORMERS_PIPELINE,

    ANALYZE_CARDIFFNLP_SENTIMENT,
    ANALYZE_CARDIFFNLP_EMOTION,
    ANALYZE_FINITEAUTOMATA_SENTIMENT,
    ANALYZE_FINITEAUTOMATA_EMOTION
)
from rebyu.pipeline.pipeline import RebyuPipeline

BLANK_PIPELINE = RebyuPipeline(
    pid='blank-pipeline',
    steps=[]
)

TEST_PIPELINE = RebyuPipeline(
    pid='test-pipeline',
    steps=[
        PREP_CAST_NAN,
        PREP_CAST_CASE,
        PREP_EXPAND_CONTRACTIONS,
        PREP_REMOVE_NUMBERS,
        PREP_REMOVE_PUNCTUATIONS,
        PREP_REMOVE_WHITESPACES,
        PREP_NLTK_TOKENIZE,
        COMPOSE_COUNTER_VOCAB,
        COMPOSE_SET_CHARVOCAB
    ]
)

TEXTBLOB_PIPELINE = RebyuPipeline(
    pid='textblob-pipeline',
    steps=[
        PREP_CAST_NAN,
        PREP_REMOVE_PUNCTUATIONS,
        PREP_SENTENCE_LENGTH,
        PREP_TEXTBLOB,
        PREP_TEXTBLOB_TOKENIZE,
        PREP_WORD_COUNT,
        COMPOSE_COUNTER_VOCAB,
        COMPOSE_COUNTER_CHARVOCAB,
        ANALYZE_TEXTBLOB_POLARITY
    ]
)
"""TextBlob-based Pipeline for Composition Extraction and Sentiment Analysis\n

1. PREP_CASE_NAN - Casting Non-Str to Empty String\n
2. PREP_REMOVE_PUNCTUATIONS - Remove Punctuations
3. PREP_SENTENCE_LENGTH - Get the Length of Text
4. PREP_TEXTBLOB - Transform to TextBlob Object
5. PREP_TEXTBLOB_TOKENIZE - Tokenize using TextBlob.tokens
6. PREP_WORD_COUNT - Get the Total Word Count
7. COMPOSE_COUNTER_VOCAB - Create a Counter Vocabulary
8. COMPOSE_COUNTER_CHARVOCAB - Create a Counter Character Vocabulary
9. ANALYZE_TEXTBLOB_POLARITY - Get the Polarity of the Text
"""

NLTK_PIPELINE = RebyuPipeline(
    pid='nltk-pipeline',
    steps=[
        PREP_CAST_NAN,
        PREP_REMOVE_PUNCTUATIONS,
        PREP_SENTENCE_LENGTH,
        PREP_NLTK_TOKENIZE,
        PREP_WORD_COUNT,
        COMPOSE_NLTK_VOCAB,
        COMPOSE_COUNTER_CHARVOCAB,
        COMPOSE_NLTK_POS_TAG,
        COMPOSE_NLTK_NER,
        ANALYZE_VADER_POLARITY
    ]
)
"""NLTK-based Pipeline for Composition Extraction and Sentiment Analysis\n

1. PREP_CASE_NAN - Casting Non-Str to Empty String\n
2. PREP_REMOVE_PUNCTUATIONS - Remove Punctuations
3. PREP_SENTENCE_LENGTH - Get the Length of Text
4. PREP_NLTK_TOKENIZE - Tokenize using NLTK
5. PREP_WORD_COUNT - Get the Total Word Count
6. COMPOSE_NLTK_VOCAB - Create an NLTK Vocabulary
7. COMPOSE_COUNTER_CHARVOCAB - Create a Counter Character Vocabulary
8. COMPOSE_NLTK_POS_TAG - Extract Part-of-Speech Tags
9. COMPOSE_NLTK_NER - Extract Named-Entities
10. ANALYZE_VADER_POLARITY - Get the Polarity of the Text
"""

CARDIFFNLP_PIPELINE = RebyuPipeline(
    pid='cardiff-nlp',
    steps=[
        PREP_CAST_NAN,
        PREP_TRIM_TEXT,
        PREP_CENSOR_USERNAME,
        PREP_CENSOR_URLS,
        PREP_REMOVE_EMOJIS,
        ANALYZE_CARDIFFNLP_SENTIMENT,
        ANALYZE_CARDIFFNLP_EMOTION
    ]
)

FINITEAUTOMATA_PIPELINE = RebyuPipeline(
    pid='finiteautomata',
    steps=[
        PREP_CAST_NAN,
        PREP_TRIM_TEXT,
        PREP_CENSOR_USERNAME,
        PREP_CENSOR_URLS,
        ANALYZE_FINITEAUTOMATA_SENTIMENT,
        ANALYZE_FINITEAUTOMATA_EMOTION
    ]
)


from typing import Any

from rebyu.util.dependency import nltk_dependency_mgt

import nltk
from nltk import pos_tag


def nltk_extract_pos_tags(series: Any):
    """ Extract Part-of-Speech Tags from a given text data using NLTK (pos_tag)

    :param series: Any series of data
    :return: Nested List of Tuples (List[List[Tuples], ...])
    """
    nltk_dependency_mgt(required=['punkt', 'averaged_perceptron_tagger'])

    pos_tags = []
    for data in series:
        if type(data) is str:
            tokens = nltk.word_tokenize(data)
        else:
            tokens = data

        pos_tags.append(pos_tag(tokens))
    return pos_tags



from typing import List

import nltk


def nltk_dependency_mgt(required: List[str] = None):
    """ Check and Fulfill required dependencies for NLTK

    :param required: List of String (packages)
    :return:
    """
    nltk_lookup = {
        'punkt': 'tokenizers/punkt',
        'stopwords': 'corpora/stopwords',
        'wordnet': 'corpora/wordnet',
        'words': 'corpora/words',
        'averaged_perceptron_tagger': 'taggers/averaged_perceptron_tagger',
        'vader_lexicon': 'sentiment/vader_lexicon',
        'maxent_ne_chunker': 'chunkers/maxent_ne_chunker'
    }

    if required is None:
        return

    for resource in required:
        try:
            nltk.data.find(nltk_lookup.get(resource, f'corpora/{resource}'))
        except LookupError:
            nltk.download(resource)
    return



from typing import Any

from rebyu.util.dependency import nltk_dependency_mgt

import nltk
from nltk import pos_tag
from nltk.chunk import ne_chunk


def nltk_extract_ner(series: Any):
    """ Extract Named-Entities from a given text data using NLTK (ne_chunk)

    :param series: Any series of data
    :return: Nested List of Tuples (List[List[Tuples], ...])
    """
    nltk_dependency_mgt(required=[
        'punkt', 'maxent_ne_chunker', 'words', 'averaged_perceptron_tagger'
    ])

    entities = []
    for data in series:
        if type(data) is str:
            tokens = nltk.word_tokenize(data)
        else:
            tokens = data

        tags = pos_tag(tokens)
        ner_chunks = ne_chunk(tags)

        named_entities = []
        for subtree in ner_chunks:
            if type(subtree) == nltk.Tree:
                entity = ' '.join([word for word, tag in subtree.leaves()])
                entity_type = subtree.label()
                named_entities.append((entity, entity_type))
        entities.append(named_entities)
    return entities



import copy

import pandas as pd

from typing import Any, AnyStr, List, Dict, Optional

from rebyu.util.logger import Logger


class BaseStep(object):
    """
    Base Class for Step

    The BaseStep class is the foundation of Rebyu Steps which are encapsulated functions to run
    within a pipeline with a given rule-set. Primarily, the step takes 3 run-types, which are preprocessor,
    composer, and analyzer.
    """
    STEP_PREPROCESS = "PREPROCESS"
    STEP_COMPOSE = "COMPOSE"
    STEP_ANALYZE = "ANALYZE"

    def __init__(self,
                 sid: AnyStr,
                 stype: AnyStr,
                 source: Any,
                 target: Any,
                 func: Any,
                 func_args: Dict[AnyStr, Any] = None):
        self.sid = sid
        self.stype = stype
        self.source = source
        self.target = target
        self.func = func
        self.func_args = func_args or {}

        self.next = None

    def run(self, data: pd.DataFrame, composition: Dict, analysis: Dict):
        """ Run the step according to it's type.

        - PREPROCESS: Takes data from Rebyu.data as source to operate and output to the target column.
        - COMPOSE: Takes data from Rebyu.data as source to operate and output to a composition dict store with target as key.
        - ANALYZE: Takes data from Rebyu.data as source to operate and output to a analysis dict store with target as key.

        :param data: The input dataset (pd.DataFrame).
        :param composition: The composition store (Dict).
        :param analysis: The analysis store (Dict).
        :return: data, composition, analysis.
        """
        if self.stype == self.STEP_PREPROCESS:
            data[self.target] = data[self.source].apply(self.func, **self.func_args)

        if self.stype == self.STEP_COMPOSE:
            composition[self.target] = self.func(data[self.source], **self.func_args)

        if self.stype == self.STEP_ANALYZE:
            analysis[self.target] = self.func(data[self.source], **self.func_args)

        return data, composition, analysis

    def copy(self, deep: bool = True):
        """ Copy the instance of this class.

        :param deep: Determines whether it will be a deepcopy.
        :return: BaseStep.
        """
        if deep:
            return copy.deepcopy(self)
        return copy.copy(self)

    def add_args(self, **kwargs):
        """ Add function arguments to BaseStep.func function.

        :param kwargs:
        :return: BaseStep.
        """
        self.func_args.update({k: v for k, v in kwargs.items()})
        return self

    def set_sid(self, sid: AnyStr):
        """ Change Step ID.

        :param sid: Step ID.
        :return: BaseStep.
        """
        self.sid = sid
        return self

    def set_stype(self, stype: AnyStr):
        """ Change Step Type.

        :param stype: Step Type (BaseStep.STEP_PREPROCESS, BaseStep.STEP_COMPOSE, BaseStep.STEP_ANALYZE).
        :return: BaseStep.
        """
        self.stype = stype
        return self

    def set_source(self, source: AnyStr):
        """ Change Source of Data (Column from Rebyu.data).

        :param source: Source Key (Rebyu.data) Key.
        :return: BaseStep.
        """
        self.source = source
        return self

    def set_target(self, target: AnyStr):
        """ Change Target of Data (Column from Rebyu.data).

        :param target: Target Key (Rebyu.data, Rebyu.composition, Rebyu.analysis) Key.
        :return: BaseStep.
        """
        self.target = target
        return self

    def set_func(self, func: Any):
        """ Change Function.

        :param func: Function Object.
        :return:
        """
        self.func = func
        return self

    def __repr__(self):
        return f'Step(sid={self.sid}, stype={self.stype}, source={self.source}, target={self.target})'


class BasePipeline(object):

    def __init__(self, pid: AnyStr, steps: List[BaseStep]):
        self.pid = pid
        self.head: Optional[BaseStep] = None
        self.tail: Optional[BaseStep] = None
        self.curr_idx, self.curr = (0, self.head)
        self.length = 0

        self.pipeline_logger = Logger(self.__class__.__name__)
        self._prepare_pipeline(steps)

    def _prepare_pipeline(self, steps: List[BaseStep]):
        """ Prepare Pipeline Steps into Linked List

        :param steps: List of Steps.
        :return:
        """
        for step in steps:
            self.add(step)

    def step(self,
             data: pd.DataFrame,
             composition: Dict,
             analysis: Dict,
             verbose: bool = False) -> bool:
        """ Take a step into the pipeline. Processing the next step to operate.

        :param data: The input dataset (pd.DataFrame).
        :param composition: The composition store (Dict).
        :param analysis: The analysis store (Dict).
        :param verbose: Determines whether verbose output is enabled.
        :return:
        """
        if self.head is None:
            return False

        if self.curr is None:
            return False

        if verbose:
            self.pipeline_logger.info(f'[{self.curr.stype}] - Running step {self.curr.sid}')

        self.curr.run(data=data, composition=composition, analysis=analysis)

        if verbose:
            self.pipeline_logger.info(f'[{self.curr.stype}] - Finished step {self.curr.sid}')

        self.curr = self.curr.next
        self.curr_idx += 1

        return True

    def run(self, data: pd.DataFrame, composition: Dict, analysis: Dict, verbose: bool = False):
        """ Run through the steps in the pipeline. Processing all the steps to operate.

        :param data: The input dataset (pd.DataFrame).
        :param composition: The composition store (Dict).
        :param analysis: The analysis store (Dict).
        :param verbose: Determines whether verbose output is enabled.
        :return:
        """
        self.pipeline_logger.info(f'[{self.pid}] - Running pipeline')
        out = True
        while out:
            out = self.step(
                data=data,
                composition=composition,
                analysis=analysis,
                verbose=verbose
            )
        self.pipeline_logger.info(f'[{self.pid}] - Finished pipeline')

    def add(self, step: BaseStep) -> BaseStep:
        """ Add a new step to the pipeline.

        :param step: BaseStep
        :return: BaseStep
        """
        new_step = step.copy()
        return self._add_node(new_step)

    def add_function(self,
                     func: object,
                     stype: AnyStr,
                     source: AnyStr,
                     target: AnyStr,
                     func_args: Dict = None) -> BaseStep:
        """ Add a new step to the pipeline via functions

        :param func: Function
        :param stype: Step Type (BaseStep.STEP_PREPROCESS, BaseStep.STEP_COMPOSE, BaseStep.STEP_ANALYZE).
        :param source: Source Key (Rebyu.data) Key.
        :param target: Target Key (Rebyu.data, Rebyu.composition, Rebyu.analysis) Key.
        :param func_args: Function arguments (Dict)
        :return: BaseStep
        """
        new_step = BaseStep(
            sid=func.__name__,
            stype=stype,
            source=source,
            target=target,
            func=func,
            func_args=func_args
        )

        return self._add_node(new_step)

    def _add_node(self, new_step: BaseStep) -> BaseStep:
        if self.head is None:
            self.head = new_step
            self.tail = new_step
            self.curr_idx, self.curr = (0, self.head)
        else:
            self.tail.next = new_step
            self.tail = new_step

        self.length += 1
        return new_step

    def pop(self) -> Optional[BaseStep]:
        """ Remove last step in the pipeline.

        :return: BaseStep | None
        """
        if self.head is None:
            return None

        if self.head == self.tail:
            step = self.head
            self.head = None
            self.tail = None
            self.length -= 1
            return step

        current = self.head
        while current.next != self.tail:
            current = current.next

        step = self.tail
        current.next = None
        self.tail = current
        self.length -= 1
        return step

    def steps_info(self) -> List[BaseStep]:
        """ Get the information of all the steps in the pipeline.

        :return: List of BaseSteps
        """
        out = []
        current = self.head
        while current:
            out.append(current)
            current = current.next
        return out

    def state(self) -> (Optional[int], Optional[BaseStep]):
        """ Get the current state of the pipeline progress.

        :return: State of Index and Current Step
        """
        return self.curr_idx, self.curr

    def reset(self):
        """ Reset the current state (Index and Step) of the pipeline.

        :return:
        """
        self.curr = self.head
        self.curr_idx = 0
        return

    def __len__(self):
        return self.length

    def __repr__(self):
        return f'Pipeline(pid={self.pid})'



from typing import AnyStr, List

from rebyu.pipeline.base import BaseStep, BasePipeline


class RebyuPipeline(BasePipeline):

    """
    RebyuPipeline

    The main class to access or initiate a Rebyu Pipeline. A pipeline is a series of steps to execute.
    A pipeline can be run each step at a time, maintaining its current step to progress in the pipeline.
    """

    def __init__(self, pid: AnyStr, steps: List[BaseStep]):
        super().__init__(pid, steps)



from typing import Any
from collections import Counter

from rebyu.util.dependency import nltk_dependency_mgt

from nltk.lm import Vocabulary


def counter_vocab(series: Any, cutoff: int = 2):
    """ Create a Vocabulary using the Counter class

    :param series: Any series of data
    :param cutoff: Minimum occurrence to enter the Vocab
    :return: collections.Counter
    """
    vocab = Counter()
    for data in series:
        vocab.update(data)

    for word in list(vocab):
        if vocab[word] < cutoff:
            del vocab[word]

    return vocab


def counter_character_vocab(series: Any, cutoff: int = 0):
    """ Create a Character Vocabulary using the Counter class

    :param series: Any series of data
    :param cutoff: Minimum occurrence to enter the Vocab
    :return: collections.Counter
    """
    vocab = Counter()
    for data in series:         # Per Row
        for unit in data:       # Per Token
            vocab.update(unit)  # Individual Token

    for char in list(vocab):
        if vocab[char] < cutoff:
            del vocab[char]

    return vocab


def set_character_vocab(series: Any):
    """ Create a Character Vocabulary using a Set

    :param series: Any series of data
    :return: set
    """
    vocab = set()
    for data in series:              # Per Row
        for unit in data:            # Per Token
            vocab.update(set(unit))  # Individual Token

    return vocab


def nltk_vocab(tokens: Any, cutoff: int = 2):
    """ Create a Vocabulary using NLTK Vocabulary class

    :param tokens: List of Tokens
    :param cutoff: Minimum occurrence to enter the vocab
    :return: nltk.lm.Vocabulary
    """
    nltk_dependency_mgt(required=['punkt'])

    tokenized_soup = [x for token in tokens for x in token]
    return Vocabulary(tokenized_soup, unk_cutoff=cutoff)



# Rebyu
# An automatic Review Analysis Toolkit
#
# Authors: Abhishta Gatya <abhishtagatya@yahoo.com>
# URL: <https://github.com/abhishtagatya/rebyu>


from rebyu.core import Rebyu
from rebyu.pipeline import step, pipeline
from rebyu.preprocess import remove, transform
from rebyu.compose import vocab
from rebyu.analysis import sentiment, misc, embedding



from typing import Any, AnyStr, List

import pandas as pd
from transformers import pipeline
from transformers import AutoModel


def transformers_model(
        series: Any,
        model: Any = AutoModel,
        model_name: AnyStr = None):
    """ Predict a given series by construct a transformers model with a set of specifications

    :param series: Any series of data
    :param model: The transformers model (from transformers)
    :param model_name: The pretrained model name
    :return: List of Prediction Object
    """
    outputs = []
    _model = model.from_pretrained(model_name)

    for data in series:
        outputs.append(_model(**data))

    return outputs


def transformers_pipeline(
        series: Any,
        task: AnyStr = 'sentiment-analysis',
        model: AnyStr = None,
        **kwargs):
    """ Predict a given series by constructing a transformers pipeline with a set of specifications

    :param series: Any series of data
    :param task: The name of the task for the pretrained model
    :param model: The pretrained model name
    :param kwargs: Additional arguments to the pipeline
    :return: List of Prediction Object
    """
    pipe_task = pipeline(task=task, model=model, **kwargs)

    outputs = []
    for data in series:
        outputs.append(pipe_task(data)[0])
    return outputs





import numpy as np
from typing import Any, Dict, Tuple, Union

POLARITY_LABEL_3 = {0: 'Negative', 1: 'Neutral', 2: 'Positive'}
POLARITY_LABEL_5 = {0: 'Negative', 1: 'Slightly Negative', 2: 'Neutral', 3: 'Slightly Positive', 4: 'Positive'}


def linear_map(value: Any, n: int = 3, label: Union[Dict, None] = None):
    """ Linearly Scale values into N Categories. Values from -1 to 1.

    :param value: Any value of data
    :param n: Number of Categories
    :param label: (Optional) Cast Unlabelled values into Labelled
    :return: Category (Labelled or Unlabelled)
    """
    interval = 2 / n
    scaled = int((value + 1) // interval)
    category = max(0, min(n-1, scaled))

    if label:
        return label.get(category, category)
    return category


def polarity_threshold_map(
        value: Any,
        threshold: Tuple[int, int] = (-0.05, 0.05),
        label: Union[Dict, None] = POLARITY_LABEL_3):
    """ Polarity Mapping by Threshold (Only 3 Categories)

    :param value: Any value of data
    :param threshold: Negative and Positive Threshold (-0.05, 0.05)
    :param label: (Optional) Cast Unlabelled values into Labelled
    :return: Category (Labelled or Unlabelled)
    """
    neg, pos = min(threshold), max(threshold)

    out = 1  # Neutral
    if value > pos:
        out += 1  # Positive
    if value < neg:
        out -= 1  # Negative

    if label:
        return label.get(out, out)
    return out
